@@ -10,12 +10,22 @@ module Minitest
   class Test < Runnable
     require "minitest/assertions"
     include Minitest::Assertions
+    include Minitest::Reportable
 
-    PASSTHROUGH_EXCEPTIONS = [NoMemoryError, SignalException, # :nodoc:
-                              Interrupt, SystemExit]
+    def class_name # :nodoc:
+      self.class.name # for Minitest::Reportable
+    end
 
+    PASSTHROUGH_EXCEPTIONS = [NoMemoryError, SignalException, SystemExit] # :nodoc:
+
+    SETUP_METHODS = %w[ before_setup setup after_setup ] # :nodoc:
+
+    TEARDOWN_METHODS = %w[ before_teardown teardown after_teardown ] # :nodoc:
+
+    # :stopdoc:
     class << self; attr_accessor :io_lock; end
     self.io_lock = Mutex.new
+    # :startdoc:
 
     ##
     # Call this at the top of your tests when you absolutely
@@ -38,9 +48,7 @@ module Minitest
     def self.make_my_diffs_pretty!
       require "pp"
 
-      define_method :mu_pp do |o|
-        o.pretty_inspect
-      end
+      define_method :mu_pp, &:pretty_inspect
     end
 
     ##
@@ -63,8 +71,8 @@ module Minitest
 
       case self.test_order
       when :random, :parallel then
-        max = methods.size
-        methods.sort.sort_by { rand max }
+        srand Minitest.seed
+        methods.sort.shuffle
       when :alpha, :sorted then
         methods.sort
       else
@@ -80,20 +88,6 @@ module Minitest
       :random
     end
 
-    ##
-    # The time it took to run this test.
-
-    attr_accessor :time
-
-    def marshal_dump # :nodoc:
-      super << self.time
-    end
-
-    def marshal_load ary # :nodoc:
-      self.time = ary.pop
-      super
-    end
-
     ##
     # Runs a single test with setup/teardown hooks.
 
@@ -101,12 +95,14 @@ module Minitest
       with_info_handler do
         time_it do
           capture_exceptions do
-            before_setup; setup; after_setup
+            SETUP_METHODS.each do |hook|
+              self.send hook
+            end
 
             self.send self.name
           end
 
-          %w{ before_teardown teardown after_teardown }.each do |hook|
+          TEARDOWN_METHODS.each do |hook|
             capture_exceptions do
               self.send hook
             end
@@ -114,7 +110,7 @@ module Minitest
         end
       end
 
-      self # per contract
+      Result.from self # per contract
     end
 
     ##
@@ -200,80 +196,55 @@ module Minitest
     end # LifecycleHooks
 
     def capture_exceptions # :nodoc:
-      begin
-        yield
-      rescue *PASSTHROUGH_EXCEPTIONS
-        raise
-      rescue Assertion => e
-        self.failures << e
-      rescue Exception => e
-        self.failures << UnexpectedError.new(e)
-      end
-    end
-
-    ##
-    # Did this run error?
-
-    def error?
-      self.failures.any? { |f| UnexpectedError === f }
-    end
-
-    ##
-    # The location identifier of this test.
-
-    def location
-      loc = " [#{self.failure.location}]" unless passed? or error?
-      "#{self.class}##{self.name}#{loc}"
-    end
-
-    ##
-    # Did this run pass?
-    #
-    # Note: skipped runs are not considered passing, but they don't
-    # cause the process to exit non-zero.
-
-    def passed?
-      not self.failure
-    end
-
-    ##
-    # Returns ".", "F", or "E" based on the result of the run.
-
-    def result_code
-      self.failure and self.failure.result_code or "."
-    end
-
-    ##
-    # Was this run skipped?
-
-    def skipped?
-      self.failure and Skip === self.failure
-    end
-
-    def time_it # :nodoc:
-      t0 = Time.now
-
       yield
-    ensure
-      self.time = Time.now - t0
+    rescue *PASSTHROUGH_EXCEPTIONS
+      raise
+    rescue Assertion => e
+      self.failures << e
+    rescue Exception => e
+      self.failures << UnexpectedError.new(sanitize_exception e)
     end
 
-    def to_s # :nodoc:
-      return location if passed? and not skipped?
+    def sanitize_exception e # :nodoc:
+      Marshal.dump e
+      e                                         # good: use as-is
+    rescue
+      neuter_exception e
+    end
 
-      failures.map { |failure|
-        "#{failure.result_label}:\n#{self.location}:\n#{failure.message}\n"
-      }.join "\n"
+    def neuter_exception e # :nodoc:
+      bt = e.backtrace
+      msg = e.message.dup
+
+      new_exception e.class, msg, bt            # e.class can be a problem...
+    rescue
+      msg.prepend "Neutered Exception #{e.class}: "
+
+      new_exception RuntimeError, msg, bt, true # but if this raises, we die
+    end
+
+    def new_exception klass, msg, bt, kill = false # :nodoc:
+      ne = klass.new msg
+      ne.set_backtrace bt
+
+      if kill then
+        ne.instance_variables.each do |v|
+          ne.remove_instance_variable v
+        end
+      end
+
+      Marshal.dump ne                           # can raise TypeError
+      ne
     end
 
     def with_info_handler &block # :nodoc:
-      t0 = Time.now
+      t0 = Minitest.clock_time
 
       handler = lambda do
-        warn "\nCurrent: %s#%s %.2fs" % [self.class, self.name, Time.now - t0]
+        warn "\nCurrent: %s#%s %.2fs" % [self.class, self.name, Minitest.clock_time - t0]
       end
 
-      self.class.on_signal "INFO", handler, &block
+      self.class.on_signal ::Minitest.info_signal, handler, &block
     end
 
     include LifecycleHooks
@@ -281,3 +252,5 @@ module Minitest
     extend Guard
   end # Test
 end
+
+require "minitest/unit" unless defined?(MiniTest) # compatibility layer only
