@@ -28,6 +28,9 @@ class TestMinitestAssertions < Minitest::Test
 
   RUBY18 = !defined? Encoding
 
+  # not included in JRuby
+  RE_LEVELS = /\(\d+ levels\) /
+
   class DummyTest
     include Minitest::Assertions
     # include Minitest::Reportable # TODO: why do I really need this?
@@ -77,6 +80,14 @@ class TestMinitestAssertions < Minitest::Test
     self.send assert_msg, expected, msg
   end
 
+  def assert_unexpected expected
+    expected = Regexp.new expected if String === expected
+
+    assert_triggered expected, Minitest::UnexpectedError do
+      yield
+    end
+  end
+
   def clean s
     s.gsub(/^ {6,10}/, "")
   end
@@ -343,7 +354,7 @@ class TestMinitestAssertions < Minitest::Test
           @@ -1,3 +1,3 @@
           -# encoding: UTF-8
           -#    valid: false
-          +# encoding: ASCII-8BIT
+          +# encoding: #{Encoding::BINARY.name}
           +#    valid: true
            "bad-utf8-\\xF1.txt"
           EOM
@@ -362,7 +373,7 @@ class TestMinitestAssertions < Minitest::Test
           @@ -1,3 +1,3 @@
           -# encoding: US-ASCII
           -#    valid: false
-          +# encoding: ASCII-8BIT
+          +# encoding: #{Encoding::BINARY.name}
           +#    valid: true
            "bad-utf8-\\xF1.txt"
           EOM
@@ -473,7 +484,10 @@ class TestMinitestAssertions < Minitest::Test
 
   def test_assert_match
     @assertion_count = 2
-    @tc.assert_match(/\w+/, "blah blah blah")
+    m = @tc.assert_match(/\w+/, "blah blah blah")
+
+    assert_kind_of MatchData, m
+    assert_equal "blah", m[0]
   end
 
   def test_assert_match_matchee_to_str
@@ -603,12 +617,117 @@ class TestMinitestAssertions < Minitest::Test
     end
   end
 
-  def test_assert_output_without_block
+  def test_assert_output_no_block
     assert_triggered "assert_output requires a block to capture output." do
       @tc.assert_output "blah"
     end
   end
 
+  def test_assert_output_nested_assert_uncaught
+    @assertion_count = 1
+
+    assert_triggered "Epic Fail!" do
+      @tc.assert_output "blah\n" do
+        puts "blah"
+        @tc.flunk
+      end
+    end
+  end
+
+  def test_assert_output_nested_raise
+    @assertion_count = 2
+
+    @tc.assert_output "blah\n" do
+      @tc.assert_raises RuntimeError do
+        puts "blah"
+        raise "boom!"
+      end
+    end
+  end
+
+  def test_assert_output_nested_raise_bad
+    @assertion_count = 0
+
+    assert_unexpected "boom!" do
+      @tc.assert_raises do            # 2) bypassed via UnexpectedError
+        @tc.assert_output "blah\n" do # 1) captures and raises UnexpectedError
+          puts "not_blah"
+          raise "boom!"
+        end
+      end
+    end
+  end
+
+  def test_assert_output_nested_raise_mismatch
+    # this test is redundant, but illustrative
+    @assertion_count = 0
+
+    assert_unexpected "boom!" do
+      @tc.assert_raises RuntimeError do # 2) bypassed via UnexpectedError
+        @tc.assert_output "blah\n" do   # 1) captures and raises UnexpectedError
+          puts "not_blah"
+          raise ArgumentError, "boom!"
+        end
+      end
+    end
+  end
+
+  def test_assert_output_nested_throw_caught
+    @assertion_count = 2
+
+    @tc.assert_output "blah\n" do
+      @tc.assert_throws :boom! do
+        puts "blah"
+        throw :boom!
+      end
+    end
+  end
+
+  def test_assert_output_nested_throw_caught_bad
+    @assertion_count = 1            # want 0; can't prevent throw from escaping :(
+
+    @tc.assert_throws :boom! do     # 2) captured via catch
+      @tc.assert_output "blah\n" do # 1) bypassed via throw
+        puts "not_blah"
+        throw :boom!
+      end
+    end
+  end
+
+  def test_assert_output_nested_throw_mismatch
+    @assertion_count = 0
+
+    assert_unexpected "uncaught throw :boom!" do
+      @tc.assert_throws :not_boom! do # 2) captured via assert_throws+rescue
+        @tc.assert_output "blah\n" do # 1) bypassed via throw
+          puts "not_blah"
+          throw :boom!
+        end
+      end
+    end
+  end
+
+  def test_assert_output_uncaught_raise
+    @assertion_count = 0
+
+    assert_unexpected "RuntimeError: boom!" do
+      @tc.assert_output "blah\n" do
+        puts "not_blah"
+        raise "boom!"
+      end
+    end
+  end
+
+  def test_assert_output_uncaught_throw
+    @assertion_count = 0
+
+    assert_unexpected "uncaught throw :boom!" do
+      @tc.assert_output "blah\n" do
+        puts "not_blah"
+        throw :boom!
+      end
+    end
+  end
   def test_assert_predicate
     @tc.assert_predicate "", :empty?
   end
@@ -643,12 +762,12 @@ class TestMinitestAssertions < Minitest::Test
       Class: <SomeError>
       Message: <\"blah\">
       ---Backtrace---
-      FILE:LINE:in \`test_assert_raises_default_triggered\'
+      FILE:LINE:in \`block in test_assert_raises_default_triggered\'
       ---------------
     EOM
 
     actual = e.message.gsub(/^.+:\d+/, "FILE:LINE")
-    actual.gsub!(/block \(\d+ levels\) in /, "") if RUBY_VERSION >= "1.9.0"
+    actual.gsub!(RE_LEVELS, "") unless jruby?
 
     assert_equal expected, actual
   end
@@ -671,11 +790,24 @@ class TestMinitestAssertions < Minitest::Test
     end
   end
 
+  def test_assert_raises_throw_nested_bad
+    @assertion_count = 0
+
+    assert_unexpected "RuntimeError: boom!" do
+      @tc.assert_raises do
+        @tc.assert_throws :blah do
+          raise "boom!"
+          throw :not_blah
+        end
+      end
+    end
+  end
+
   ##
   # *sigh* This is quite an odd scenario, but it is from real (albeit
   # ugly) test code in ruby-core:
 
-  # http://svn.ruby-lang.org/cgi-bin/viewvc.cgi?view=rev&revision=29259
+  # https://svn.ruby-lang.org/cgi-bin/viewvc.cgi?view=rev&revision=29259
 
   def test_assert_raises_skip
     @assertion_count = 0
@@ -709,12 +841,12 @@ class TestMinitestAssertions < Minitest::Test
       Class: <AnError>
       Message: <\"some message\">
       ---Backtrace---
-      FILE:LINE:in \`test_assert_raises_subclass_triggered\'
+      FILE:LINE:in \`block in test_assert_raises_subclass_triggered\'
       ---------------
     EOM
 
     actual = e.message.gsub(/^.+:\d+/, "FILE:LINE")
-    actual.gsub!(/block \(\d+ levels\) in /, "") if RUBY_VERSION >= "1.9.0"
+    actual.gsub!(RE_LEVELS, "") unless jruby?
 
     assert_equal expected.chomp, actual
   end
@@ -731,12 +863,12 @@ class TestMinitestAssertions < Minitest::Test
       Class: <SyntaxError>
       Message: <\"icky\">
       ---Backtrace---
-      FILE:LINE:in \`test_assert_raises_triggered_different\'
+      FILE:LINE:in \`block in test_assert_raises_triggered_different\'
       ---------------
     EOM
 
     actual = e.message.gsub(/^.+:\d+/, "FILE:LINE")
-    actual.gsub!(/block \(\d+ levels\) in /, "") if RUBY_VERSION >= "1.9.0"
+    actual.gsub!(RE_LEVELS, "") unless jruby?
 
     assert_equal expected, actual
   end
@@ -754,12 +886,12 @@ class TestMinitestAssertions < Minitest::Test
       Class: <SyntaxError>
       Message: <\"icky\">
       ---Backtrace---
-      FILE:LINE:in \`test_assert_raises_triggered_different_msg\'
+      FILE:LINE:in \`block in test_assert_raises_triggered_different_msg\'
       ---------------
     EOM
 
     actual = e.message.gsub(/^.+:\d+/, "FILE:LINE")
-    actual.gsub!(/block \(\d+ levels\) in /, "") if RUBY_VERSION >= "1.9.0"
+    actual.gsub!(RE_LEVELS, "") unless jruby?
 
     assert_equal expected.chomp, actual
   end
@@ -869,13 +1001,25 @@ class TestMinitestAssertions < Minitest::Test
   end
 
   def test_assert_throws
-    @tc.assert_throws :blah do
+    v = @tc.assert_throws :blah do
       throw :blah
     end
+
+    assert_nil v
+  end
+
+  def test_assert_throws_value
+    v = @tc.assert_throws :blah do
+      throw :blah, 42
+    end
+
+    assert_equal 42, v
   end
 
   def test_assert_throws_argument_exception
-    @tc.assert_raises ArgumentError do
+    @assertion_count = 0
+
+    assert_unexpected "ArgumentError" do
       @tc.assert_throws :blah do
         raise ArgumentError
       end
@@ -891,7 +1035,9 @@ class TestMinitestAssertions < Minitest::Test
   end
 
   def test_assert_throws_name_error
-    @tc.assert_raises NameError do
+    @assertion_count = 0
+
+    assert_unexpected "NameError" do
       @tc.assert_throws :blah do
         raise NameError
       end
@@ -988,16 +1134,20 @@ class TestMinitestAssertions < Minitest::Test
     end
   end
 
+  def assert_fail_after t
+    @tc.fail_after t.year, t.month, t.day, "remove the deprecations"
+  end
+
   def test_fail_after
-    t = Time.now
-    y, m, d = t.year, t.month, t.day
+    d0 = Time.now
+    d1 = d0 + 86_400 # I am an idiot
 
     assert_silent do
-      @tc.fail_after y, m, d+1, "remove the deprecations"
+      assert_fail_after d1
     end
 
     assert_triggered "remove the deprecations" do
-      @tc.fail_after y, m, d, "remove the deprecations"
+      assert_fail_after d0
     end
   end
 
@@ -1020,7 +1170,7 @@ class TestMinitestAssertions < Minitest::Test
   def test_refute
     @assertion_count = 2
 
-    @tc.assert_equal false, @tc.refute(false), "returns false on success"
+    @tc.assert_equal true, @tc.refute(false), "returns true on success"
   end
 
   def test_refute_empty
@@ -1212,18 +1362,22 @@ class TestMinitestAssertions < Minitest::Test
     end
   end
 
+  def assert_skip_until t, msg
+    @tc.skip_until t.year, t.month, t.day, msg
+  end
+
   def test_skip_until
     @assertion_count = 0
 
-    t = Time.now
-    y, m, d = t.year, t.month, t.day
+    d0 = Time.now
+    d1 = d0 + 86_400 # I am an idiot
 
     assert_output "", /Stale skip_until \"not yet\" at .*?:\d+$/ do
-      @tc.skip_until y, m, d, "not yet"
+      assert_skip_until d0, "not yet"
     end
 
     assert_triggered "not ready yet", Minitest::Skip do
-      @tc.skip_until y, m, d+1, "not ready yet"
+      assert_skip_until d1, "not ready yet"
     end
   end
 
@@ -1377,14 +1531,14 @@ class TestMinitestAssertionHelpers < Minitest::Test
 
   def test_mu_pp_for_diff_str_encoding
     str = "A\nB".b
-    exp = "# encoding: ASCII-8BIT\n#    valid: true\n\"A\nB\""
+    exp = "# encoding: #{Encoding::BINARY.name}\n#    valid: true\n\"A\nB\""
 
     assert_mu_pp_for_diff exp, str, :raw
   end
 
   def test_mu_pp_for_diff_str_encoding_both
     str = "A\\n\nB".b
-    exp = "# encoding: ASCII-8BIT\n#    valid: true\n\"A\\\\n\\nB\""
+    exp = "# encoding: #{Encoding::BINARY.name}\n#    valid: true\n\"A\\\\n\\nB\""
 
     assert_mu_pp_for_diff exp, str, :raw
   end
@@ -1424,7 +1578,7 @@ class TestMinitestAssertionHelpers < Minitest::Test
 
   def test_mu_pp_str_encoding
     str = "A\nB".b
-    exp = "# encoding: ASCII-8BIT\n#    valid: true\n\"A\\nB\""
+    exp = "# encoding: #{Encoding::BINARY.name}\n#    valid: true\n\"A\\nB\""
 
     assert_mu_pp exp, str, :raw
   end
